import json
from typing import Optional, Text, List, Dict, Any
from unittest.mock import patch, Mock
from uuid import uuid4

import pytest
import time
from _pytest.monkeypatch import MonkeyPatch
from _pytest.tmpdir import TempdirFactory
from sqlalchemy.orm import Session

import rasax.community.services.event_consumers.utils as consumer_utils
import rasax.community.utils as rasa_x_utils
import tests.unit.services.test_data_service as data_test
from rasa.utils.endpoints import EndpointConfig
from rasax.community import config, telemetry
from rasax.community.config import team_name
from rasax.community.constants import SHARE_YOUR_BOT_CHANNEL_NAME
from rasax.community.database.conversation import Conversation
from rasax.community.services.event_consumers.kafka_consumer import KafkaEventConsumer
from rasax.community.services.event_consumers.pika_consumer import PikaEventConsumer
from rasax.community.services.event_consumers.sqlite_consumer import SQLiteEventConsumer
from rasax.community.services.event_service import (
    EventService,
    read_event_broker_config,
)
from rasax.community.services.intent_service import (
    INTENT_NAME_KEY,
    INTENT_MAPPED_TO_KEY,
    INTENT_EXAMPLES_KEY,
)
from rasax.community.services.logs_service import LogsService
from rasax.community.services.model_service import ModelService
from tests.unit.api.blueprints.conftest import (
    generate_conversation,
    tag_1,
    get_conversation_by_id,
    tag_2,
)
from tests.unit.conftest import (
    user_event,
    insert_test_model,
    insert_test_user,
    action_event,
    bot_event,
)


def get_tag_by_value(tag_value: int, tags: List[Dict[Text, Any]]) -> Dict[Text, Any]:
    for tag in tags:
        if tag["value"] == tag_value:
            return tag
    raise ValueError(f"Tag with value: {tag_value} was not found")


def test_get_sender_ids(event_service: EventService):
    assert event_service.sender_ids() == []

    events = [
        user_event("test_get_sender_ids1"),
        user_event("test_get_sender_ids2"),
        user_event("test_get_sender_ids1"),
    ]

    for event in events:
        event_service.save_event(json.dumps(event))

    assert event_service.sender_ids() == [
        "test_get_sender_ids1",
        "test_get_sender_ids2",
    ]


def test_save_user_event(event_service: EventService):
    test_sender = "test save sender"
    test_timestamp = 123456.03423423
    test_intent = "test save intent"
    test_event = user_event(test_sender, test_timestamp, test_intent)
    test_event_string = json.dumps(test_event)
    event_service.save_event(test_event_string)

    expected = {
        "sender_id": test_sender,
        "sender_name": test_sender,
        "latest_event_time": test_timestamp,
        "intents": [test_intent],
        "actions": [],
        "latest_input_channel": "rest",
        "minimum_action_confidence": 1,
        "in_training_data": True,
        "policies": [],
        "n_user_messages": 1,
        "corrected_messages": [],
        "has_flagged_messages": False,
        "interactive": False,
    }

    actual = event_service.get_conversation_metadata_for_client(test_sender)

    assert actual == expected

    # Save second event and check whether conversation is updated correctly.
    test_intent2 = "test save intent 2"
    test_event["parse_data"]["intent"]["name"] = test_intent2
    test_event["input_channel"] = "rest2"
    test_event["timestamp"] = 223456

    event_service.save_event(json.dumps(test_event))

    expected["intents"] = [test_intent, test_intent2]
    expected["latest_input_channel"] = test_event["input_channel"]
    expected["n_user_messages"] = 2
    expected["latest_event_time"] = test_event["timestamp"]

    actual = event_service.get_conversation_metadata_for_client(test_sender)

    assert actual == expected


def test_save_action_event(event_service: EventService):
    event = action_event(
        "test save action sender", 323456, "test save action", "test save policy", 0.1
    )

    event_service.save_event(json.dumps(event))

    expected = {
        "sender_id": event["sender_id"],
        "sender_name": event["sender_id"],
        "latest_event_time": event["timestamp"],
        "intents": [],
        "actions": [event["name"]],
        "latest_input_channel": None,
        "minimum_action_confidence": event["confidence"],
        "in_training_data": False,
        "policies": [event["policy"]],
        "n_user_messages": 0,
        "corrected_messages": [],
        "has_flagged_messages": False,
        "interactive": False,
    }

    actual = event_service.get_conversation_metadata_for_client(event["sender_id"])

    assert actual == expected

    # Save second event and check whether conversation is updated correctly.
    event["confidence"] = 0.0
    event["policy"] = "test save policy 2"

    event_service.save_event(json.dumps(event))

    expected = expected
    expected["minimum_action_confidence"] = event["confidence"]
    expected["policies"] = ["test save policy", event["policy"]]

    actual = event_service.get_conversation_metadata_for_client(event["sender_id"])

    assert actual == expected


def test_get_statistics(event_service: EventService):
    # use statistics of the events in the two tests above
    actual = event_service.get_statistics()
    expected = {
        "user_messages": 5,
        "bot_messages": 0,
        "top_intents": ["test", "test save intent", "test save intent 2"],
        "top_actions": ["test save action"],
        "top_entities": [],
        "top_policies": ["test save policy", "test save policy 2"],
    }
    assert actual == expected

    # insert another user intent with intent to see whether ordering changes
    test_event = user_event("other user", 99999, "test save intent 2")
    event_service.save_event(json.dumps(test_event))
    expected["user_messages"] = 6
    expected["top_intents"] = ["test", "test save intent 2", "test save intent"]

    actual = event_service.get_statistics()

    assert actual == expected


def test_get_empty_statistics(event_service: EventService):
    non_empty_statistics = event_service.get_statistics()

    assert event_service.user_statistics_dict().keys() == non_empty_statistics.keys()


def test_save_evaluation(event_service: EventService):
    test_evaluation = {
        "accuracy": 0.9,
        "f1": 0.9333333333333333,
        "in_training_data_fraction": 0.8571428571428571,
        "precision": 1,
        "is_end_to_end_evaluation": True,
        "actions": [
            {
                "action": "utter_ask_howcanhelp",
                "confidence": 1,
                "policy": "policy_0_MemoizationPolicy",
                "predicted": "utter_ask_howcanhelp",
            }
        ],
    }
    test_sender = "evaluation sender"

    # Create conversation for user
    event_service.save_event(json.dumps(user_event(test_sender, 0)))

    event_service.update_evaluation(test_sender, test_evaluation)

    assert event_service.get_evaluation(test_sender) == test_evaluation


def test_save_evaluation_for_invalid_conversation(event_service: EventService):
    with pytest.raises(ValueError):
        event_service.update_evaluation(
            "test_save_evaluation_for_invalid_conversation", {}
        )


def test_delete_evaluation(event_service: EventService):
    test_sender = "evaluation delete sender"
    # Create conversation for user
    event_service.save_event(json.dumps(user_event(test_sender, 0)))

    event_service.update_evaluation(test_sender, {"eval": "test"})

    assert event_service.get_evaluation(test_sender)

    event_service.delete_evaluation(test_sender)

    assert not event_service.get_evaluation(test_sender)


def test_get_unique_intents(event_service: EventService):
    expected = event_service.get_unique_intents()
    test_intent = "unique intent 1"
    event_service.save_event(json.dumps(user_event("sender", 0, test_intent)))

    expected.append(test_intent)
    actual = event_service.get_unique_intents()
    assert len(actual) == len(expected)
    assert all([e in actual for e in expected])

    event_service.save_event(json.dumps(user_event("sender", 0, test_intent)))
    actual = event_service.get_unique_intents()

    assert len(actual) == len(expected)
    assert all([e in actual for e in expected])

    test_intent2 = "unique intent 2"
    expected.append(test_intent2)
    event_service.save_event(json.dumps(user_event("sender", 0, test_intent2)))
    actual = event_service.get_unique_intents()

    assert len(actual) == len(expected)
    assert all([e in actual for e in expected])


def test_get_unique_policies(event_service: EventService):
    expected = event_service.get_unique_policies()

    test_policy = "unique policy 1"
    event_service.save_event(
        json.dumps(action_event("unique sender", policy=test_policy))
    )
    expected.append(test_policy)
    actual = event_service.get_unique_policies()
    assert len(actual) == len(expected)
    assert all([e in actual for e in expected])

    event_service.save_event(
        json.dumps(action_event("unique sender", policy=test_policy))
    )
    actual = event_service.get_unique_policies()

    assert len(actual) == len(expected)
    assert all([e in actual for e in expected])

    test_policy2 = "unique policy 2"
    expected.append(test_policy2)
    event_service.save_event(
        json.dumps(action_event("unique sender", policy=test_policy2))
    )
    actual = event_service.get_unique_policies()

    assert len(actual) == len(expected)
    assert all([e in actual for e in expected])


def test_get_unique_actions(event_service: EventService):
    expected = event_service.get_unique_actions()

    test_action = "unique action 1"
    event_service.save_event(
        json.dumps(action_event("unique sender", action_name=test_action))
    )
    expected.append(test_action)
    actual = event_service.get_unique_actions()
    assert len(actual) == len(expected)
    assert all([e in actual for e in expected])

    event_service.save_event(
        json.dumps(action_event("unique sender", action_name=test_action))
    )
    actual = event_service.get_unique_actions()

    assert len(actual) == len(expected)
    assert all([e in actual for e in expected])

    test_action2 = "unique action 2"
    expected.append(test_action2)
    event_service.save_event(
        json.dumps(action_event("unique sender", action_name=test_action2))
    )
    actual = event_service.get_unique_actions()

    assert len(actual) == len(expected)
    assert all([e in actual for e in expected])


def test_get_unique_entities(event_service: EventService):
    _user_event = user_event(
        "test_get_unique_entities1",
        entities=[{"entity": "entity1", "value": 2}, {"entity": "entity2", "value": 2}],
    )
    event_service.save_event(json.dumps(_user_event))

    _user_event = user_event(
        "test_get_unique_entities2",
        entities=[{"entity": "entity1", "value": 2}, {"entity": "entity3", "value": 2}],
    )
    event_service.save_event(json.dumps(_user_event))

    assert event_service.get_unique_entities() == ["entity1", "entity2", "entity3"]


def test_add_flag_without_conversation(event_service):
    test_sender_id = uuid4().hex
    event_service.add_flagged_message(test_sender_id, 0)

    assert not event_service.get_conversation_metadata_for_client(test_sender_id)


def test_no_initial_flag(event_service: EventService):
    test_sender_id = uuid4().hex
    test_event = action_event(test_sender_id)
    event_service.save_event(json.dumps(test_event))

    actual = event_service.get_conversation_metadata_for_client(test_sender_id)

    assert not actual.get("has_flagged_messages", None)


def test_add_flagged_message(event_service):
    test_sender_id = uuid4().hex
    test_timestamp = 5
    test_event = user_event(test_sender_id, test_timestamp)
    event_service.save_event(json.dumps(test_event))

    event_service.add_flagged_message(test_sender_id, test_timestamp)

    actual = event_service.get_conversation_metadata_for_client(test_sender_id)

    assert actual.get("has_flagged_messages", None)


def test_remove_flagged_conversation(event_service):
    test_sender_id = uuid4().hex
    test_timestamp = 5
    test_event = user_event(test_sender_id, test_timestamp)
    event_service.save_event(json.dumps(test_event))
    event_service.add_flagged_message(test_sender_id, test_timestamp)

    test_timestamp2 = 10
    test_event = user_event(test_sender_id, test_timestamp2)
    event_service.save_event(json.dumps(test_event))
    event_service.add_flagged_message(test_sender_id, test_timestamp2)

    event_service.delete_flagged_message(test_sender_id, test_timestamp)

    actual = event_service.get_conversation_metadata_for_client(test_sender_id)

    assert actual.get("has_flagged_messages", None)


def test_get_all(event_service, session):
    # Remove old conversation from old tests
    old_conversations = session.query(Conversation).all()
    event_service.delete_all(old_conversations)
    session.commit()

    test_event1 = user_event(uuid4().hex, 0)
    event_service.save_event(json.dumps(test_event1))

    test_event2 = user_event(uuid4().hex, 0)
    event_service.save_event(json.dumps(test_event2))

    _, total_conversations = event_service.get_conversation_metadata_for_all_clients()

    assert total_conversations == 2


def test_get_all_with_text_query(
    event_service: EventService, logs_service: LogsService, model_service: ModelService
):
    test_sender_id = uuid4().hex
    test_text = "test with text query"
    test_project = "default"
    test_event = json.dumps(
        user_event(test_sender_id, project=test_project, text=test_text)
    )

    # insert model for logs
    insert_test_model(test_project, "a model", model_service)

    event = event_service.save_event(test_event)
    logs_service.save_nlu_logs_from_event(test_event, event.id)

    (
        actual,
        total_conversations,
    ) = event_service.get_conversation_metadata_for_all_clients(text_query="text")

    assert total_conversations == 1
    assert actual[0]["sender_id"] == test_sender_id


def test_get_all_with_entity_query(event_service: EventService):
    test_sender_id = uuid4().hex
    test_text = "test with text query"
    test_project = "default"
    test_event = user_event(
        test_sender_id,
        project=test_project,
        text=test_text,
        entities=[{"entity": "package_manager"}],
    )
    event_service.save_event(json.dumps(test_event))

    (
        actual,
        total_conversations,
    ) = event_service.get_conversation_metadata_for_all_clients(
        entity_query="package_manager"
    )

    assert total_conversations == 1
    assert actual[0]["sender_id"] == test_sender_id


def test_get_all_with_intent_query(event_service: EventService):
    test_sender_id = uuid4().hex
    test_text = "test with text query"
    test_project = "default"
    intent = "245i2345p3i4o5"
    test_event = user_event(
        test_sender_id, project=test_project, text=test_text, intent=intent
    )

    event_service.save_event(json.dumps(test_event))

    (
        actual,
        total_conversations,
    ) = event_service.get_conversation_metadata_for_all_clients(intent_query=intent)

    assert total_conversations == 1
    assert actual[0]["sender_id"] == test_sender_id


def test_get_all_with_action_query(event_service: EventService):
    test_sender_id = uuid4().hex
    test_action = "test_get_all_with_action_query"
    test_event = action_event(test_sender_id, action_name=test_action)

    event_service.save_event(json.dumps(test_event))

    (
        actual,
        total_conversations,
    ) = event_service.get_conversation_metadata_for_all_clients(
        action_query=test_action
    )

    assert total_conversations == 1
    assert actual[0]["sender_id"] == test_sender_id


def test_get_all_with_policy_query(event_service: EventService):
    test_sender_id = uuid4().hex
    test_policy = "special-policy"
    test_event = action_event(test_sender_id, policy=test_policy)

    event_service.save_event(json.dumps(test_event))

    (
        actual,
        total_conversations,
    ) = event_service.get_conversation_metadata_for_all_clients(policies=test_policy)

    assert total_conversations == 1
    assert actual[0]["sender_id"] == test_sender_id


def test_get_only_flagged(event_service):
    test_sender_id = uuid4().hex
    test_event1 = user_event(test_sender_id, 0)
    event_service.save_event(json.dumps(test_event1))
    event_service.add_flagged_message(test_sender_id, 0)

    test_event2 = user_event("another user", 0)
    event_service.save_event(json.dumps(test_event2))

    _, total_conversations = event_service.get_conversation_metadata_for_all_clients(
        only_flagged=True
    )

    assert total_conversations == 1


def test_empty_initial_correct_messages(event_service):
    test_sender_id = uuid4().hex
    test_event = action_event(test_sender_id)
    event_service.save_event(json.dumps(test_event))

    actual = event_service.get_conversation_metadata_for_client(test_sender_id)

    assert actual.get("corrected_messages", None) == []


def test_correct_message(event_service, data_service, session):
    test_sender_id = uuid4().hex
    insert_test_user(session, test_sender_id)
    test_timestamp = 5
    test_intent = "intent"
    test_event = user_event(test_sender_id, test_timestamp, test_intent)
    # Insert user event and create conversations metadata for conversation
    event_service.save_event(json.dumps(test_event))

    # Correct message
    test_project = config.project_name
    test_user = {"username": test_sender_id, "team": team_name}
    event_service.correct_message(
        test_sender_id, test_timestamp, {"intent": test_intent}, test_user, test_project
    )

    # Insert another message into the conversation
    test_intent2 = "intent2"
    test_timestamp2 = 10
    test_event = user_event(test_sender_id, test_timestamp2, test_intent2)

    # Add permanent intent by inserting a training example
    permanent_intent = "permanent_intent"

    training_data = data_test.TEST_EXAMPLE.copy()
    training_data["intent"] = permanent_intent
    data_service.save_example(test_sender_id, test_project, training_data)

    event_service.save_event(json.dumps(test_event))

    # Correct message with an intent that is mapped to the permanent intent
    event_service.correct_message(
        test_sender_id,
        test_timestamp2,
        {INTENT_NAME_KEY: test_intent2, INTENT_MAPPED_TO_KEY: permanent_intent},
        test_user,
        test_project,
    )

    actual = event_service.get_conversation_metadata_for_client(test_sender_id)

    # Check whether messages were corrected.
    assert actual.get("corrected_messages", None) == [
        {"message_timestamp": test_timestamp, "intent": test_intent},
        {"message_timestamp": test_timestamp2, "intent": test_intent2},
    ]


def test_delete_correct_message(event_service, data_service, intent_service, session):
    # delete old temporary intents
    test_sender_id = uuid4().hex
    insert_test_user(session, test_sender_id)
    test_timestamp = 5
    test_intent = "intent"
    test_event = user_event(test_sender_id, test_timestamp, test_intent)
    # Insert user event and create conversations metadata for conversation
    event_service.save_event(json.dumps(test_event))

    # add permanent intent by inserting a training example
    permanent_intent = "permanent intent for delete message correction"
    training_data = data_test.TEST_EXAMPLE.copy()
    training_data["intent"] = permanent_intent
    data_service.save_example(test_sender_id, config.project_name, training_data)

    # Correct message
    test_user = {"username": test_sender_id, "team": team_name}
    correct_intent = {
        INTENT_NAME_KEY: test_intent,
        INTENT_MAPPED_TO_KEY: permanent_intent,
    }
    event_service.correct_message(
        test_sender_id, test_timestamp, correct_intent, test_user, config.project_name
    )
    # Delete correction
    event_service.delete_message_correction(
        test_sender_id, test_timestamp, config.project_name
    )

    actual = event_service.get_conversation_metadata_for_client(test_sender_id)

    # Assert message correction was deleted
    assert actual.get("corrected_messages") == []

    training_data, _ = data_service.get_training_data(config.project_name)

    assert len([r for r in training_data if r["intent"] == permanent_intent]) == 1

    assert (
        len(
            intent_service.get_temporary_intent(test_intent, config.project_name)[
                INTENT_EXAMPLES_KEY
            ]
        )
        == 0
    )


def test_create_interactive_learning_conversation_with_random_sender_id(
    event_service: EventService,
):
    _dict = {"interactive": True}

    actual = event_service.create_conversation_from(_dict)

    assert actual["interactive"] is True
    assert actual["sender_id"]


def test_create_interactive_learning_conversation_with_sender_id(
    event_service: EventService,
):
    _dict = {"interactive": True, "sender_id": "some sender"}

    actual = event_service.create_conversation_from(_dict)

    assert actual["interactive"] is True
    assert actual["sender_id"] == _dict["sender_id"]


def test_create_interactive_learning_conversation_with_existing_sender_id(
    event_service: EventService, session: Session
):
    _dict = {"interactive": True, "sender_id": "test_create_with_existing"}

    _ = event_service.create_conversation_from(_dict)

    with pytest.raises(ValueError):
        event_service.create_conversation_from(_dict)

    # Rollback db error for further tests
    session.rollback()


def test_get_all_without_interactive(event_service: EventService):
    _dict = {"interactive": True, "sender_id": "get_without_interactive"}
    _ = event_service.create_conversation_from(_dict)

    actual, _ = event_service.get_conversation_metadata_for_all_clients()

    assert all([c["sender_id"] != _dict["sender_id"] for c in actual])


def test_get_all_with_interactive(event_service: EventService):
    _dict = {"interactive": True, "sender_id": "get_with_interactive"}
    _ = event_service.create_conversation_from(_dict)

    actual, _ = event_service.get_conversation_metadata_for_all_clients(
        include_interactive=True
    )

    assert any([c["sender_id"] == _dict["sender_id"] for c in actual])


def test_modify_event_time_to_be_larger_than():
    test_sender = "test sender"
    timestamps = [100, 5, 120]
    events = [action_event(test_sender, t) for t in timestamps]

    minimal_timestamp = 1000

    actual = EventService.modify_event_time_to_be_later_than(minimal_timestamp, events)

    actual_timestamps = [event["timestamp"] for event in actual]

    assert actual_timestamps == [1000.001, 1000.002, 1000.003]


def test_modify_event_time_with_if_no_modification_needed():
    test_sender = "test sender"
    timestamps = [5, 100, 120]
    events = [action_event(test_sender, t) for t in timestamps]

    minimal_timestamp = 0

    actual = EventService.modify_event_time_to_be_later_than(minimal_timestamp, events)

    actual_timestamps = [event["timestamp"] for event in actual]

    assert actual_timestamps == [5, 100, 120]


def test_modify_event_time_with_empty_list():
    actual = EventService.modify_event_time_to_be_later_than(0, [])

    assert actual == []


def test_get_tracker(event_service: EventService):
    test_sender = "test get tracker"

    events = [action_event(test_sender, t) for t in [1, 2, 3]]
    events.append(action_event(test_sender, 4, action_name="action_listen"))
    events += (user_event(test_sender, t, text=str(t)) for t in [5, 6, 7])

    for event in events:
        event_service.save_event(json.dumps(event))

    actual = event_service.get_tracker_for_conversation(test_sender)

    assert len(actual.events) == 7
    assert actual.latest_message.text == "7"


def test_get_tracker_if_sender_id_not_found(event_service: EventService):
    not_existing_sender = "test_get_tracker_if_sender_id_not_found"

    assert not event_service.get_tracker_for_conversation(not_existing_sender)


def test_tracker_with_since(event_service: EventService):
    test_sender = "test_tracker_with_since"
    events = [action_event(test_sender, t) for t in [1, 2, 3, 3.01]]

    for event in events:
        event_service.save_event(json.dumps(event))

    actual = event_service.get_tracker_for_conversation(test_sender, since_time=3)

    assert len(actual.events) == 1


def test_get_tracker_with_zero_events(event_service: EventService):
    test_sender = "test_get_tracker_with_zero_events"
    events = [action_event(test_sender, t) for t in [1, 2, 3]]

    for event in events:
        event_service.save_event(json.dumps(event))

    actual = event_service.get_tracker_for_conversation(test_sender, since_time=3)

    assert list(actual.events) == []


def test_get_story_for_sender_id(event_service: EventService):
    test_sender = "test_get_story_for_sender_id"

    events = [action_event(test_sender, 1, action_name="action_listen")]
    events.append(user_event(test_sender, 2))
    events += [action_event(test_sender, t) for t in [3, 5, 6]]

    for event in events:
        event_service.save_event(json.dumps(event))

    expected = """## test_get_story_for_sender_id
* test
    - action
    - action
    - action
"""

    assert event_service.story_for_sender_id(test_sender) == expected


def test_get_tracker_including_rasax_metadata(event_service: EventService):
    test_sender = "test_get_tracker_including_rasax_metadata"
    events = [action_event(test_sender, t) for t in [1, 2, 3]]

    for event in events:
        event_service.save_event(json.dumps(event))

    event_service.add_flagged_message(test_sender, 2)

    actual = event_service.get_tracker_with_message_flags(test_sender)

    assert all(e["is_flagged"] == (e["timestamp"] == 2) for e in actual["events"])
    assert all(
        e["metadata"]["rasa_x_flagged"] == (e["timestamp"] == 2)
        for e in actual["events"]
    )


def test_get_tracker_including_rasax_metadata_with_until_time(
    event_service: EventService,
):
    test_sender = "test_get_tracker_including_rasax_metadata_with_until_time"
    events = [action_event(test_sender, t) for t in [1, 2, 3]]

    for event in events:
        event_service.save_event(json.dumps(event))

    actual = event_service.get_tracker_with_message_flags(test_sender, until_time=1)

    assert len(actual["events"]) == 1


def test_get_tracker_including_rasax_metadata_ids(event_service: EventService):
    test_sender = "test_get_tracker_including_rasax_metadata_ids"
    events = [action_event(test_sender, t) for t in range(10)]

    for event in events:
        event_service.save_event(json.dumps(event))

    actual = event_service.get_tracker_with_message_flags(test_sender)
    db_events = event_service.get_conversation_events(test_sender)

    assert db_events
    assert [e["metadata"]["rasa_x_id"] for e in actual["events"]] == [
        e.id for e in db_events
    ]


def test_get_tracker_with_correct_order(event_service: EventService):
    test_sender = "test_get_tracker_with_correct_order"
    events = [action_event(test_sender, t) for t in [10, 9, 8, 7, 11, 1]]

    for event in events:
        event_service.save_event(json.dumps(event))

    actual = event_service.get_tracker_with_message_flags(test_sender, until_time=10)
    actual_timestamps = [e["timestamp"] for e in actual["events"]]

    assert actual_timestamps == [1, 7, 8, 9, 10]


def test_get_messages_with_invalid_sender_id(event_service: EventService):
    test_sender = "test_get_messages_with_invalid_sender_id"

    assert event_service.get_messages_for(test_sender) is None


def test_get_messages_with_no_messages(event_service: EventService):
    test_sender = "test_get_messages_with_no_messages"
    events = [action_event(test_sender, t) for t in [1, 2, 3]]

    for event in events:
        event_service.save_event(json.dumps(event))

    actual = event_service.get_messages_for(test_sender)

    assert len(actual["events"]) == 3
    assert actual["messages"] == []


def test_get_messages_with_messages(event_service: EventService):
    test_sender = "test_get_messages_with_messages"
    events = [user_event(test_sender, t) for t in [1, 2, 3]]
    events += [action_event(test_sender, t) for t in [4, 5, 6]]
    events += [bot_event(test_sender, t) for t in [7, 8, 9]]

    for event in events:
        event_service.save_event(json.dumps(event))

    event_service.add_flagged_message(test_sender, 2)

    actual = event_service.get_messages_for(test_sender)

    assert all(e["is_flagged"] == (e["timestamp"] == 2) for e in actual["events"])
    assert len(actual["messages"]) == 6
    assert len(actual["events"]) == 9


def test_get_sender_name(event_service: EventService):
    test_sender = "test_get_sender_name"
    # Create conversation by adding events
    event = user_event(test_sender, 1)
    event_service.save_event(json.dumps(event))

    # noinspection PyProtectedMember
    conversation = event_service._get_conversation(test_sender)

    assert EventService.get_sender_name(conversation) == conversation.sender_id


def test_get_sender_name_with_tester_channel(event_service: EventService):
    test_sender = "test_get_sender_name_with_tester_channel"
    # Create conversation by adding events
    event = user_event(test_sender, 1)
    event_service.save_event(json.dumps(event))

    # noinspection PyProtectedMember
    conversation = event_service._get_conversation(test_sender)
    conversation.latest_input_channel = SHARE_YOUR_BOT_CHANNEL_NAME

    assert EventService.get_sender_name(conversation) == SHARE_YOUR_BOT_CHANNEL_NAME


def test_unacceptable_event_consumer():
    import rasax.community.services.event_service as event_service_module

    consumer_config = EndpointConfig(store_type="non-existing-type")

    # unacceptable consumer raises sys exit
    with pytest.raises(SystemExit):
        event_service_module.continuously_consume(consumer_config)


def liveness_endpoint_is_healthy(
    retries: int, time_between_retries_in_seconds: Optional[int] = 1
) -> bool:
    """Check whether the liveness endpoint is reachable.

    Args:
        retries: Number of retries.
        time_between_retries_in_seconds: Time between retries in seconds.

    Returns:
        `True` if endpoint is reachable, `False` if it isn't.

    """
    import requests

    for _ in range(retries):
        try:
            response = requests.get("http://0.0.0.0:5673/health")
            return response.status_code == 200
        except requests.exceptions.RequestException:
            time.sleep(time_between_retries_in_seconds)

    return False


def test_event_consumer_liveness_endpoint():
    from rasax.community.services.event_consumers.event_consumer import EventConsumer

    # noinspection PyProtectedMember
    process = EventConsumer._run_liveness_endpoint_process("test consumer")

    try:
        assert liveness_endpoint_is_healthy(retries=5)
    except AssertionError as e:
        raise e
    finally:  # kill process no matter what
        process.terminate()


def test_event_consumer_liveness_endpoint_destruction():
    # this test ensures that the endpoint is not available after an error
    # occurs in the consume method

    # mocked consume method that just raises an exception
    def mocked_consume(_):
        raise ValueError

    # noinspection PyBroadException
    try:
        with patch.object(SQLiteEventConsumer, "consume", mocked_consume):
            # noinspection PyTypeChecker
            consumer = SQLiteEventConsumer(should_run_liveness_endpoint=True)
            assert liveness_endpoint_is_healthy(retries=5)
            consumer.consume()
    except ValueError:
        consumer.kill_liveness_endpoint_process()
        assert not liveness_endpoint_is_healthy(retries=1)
    finally:  # kill process no matter what
        consumer.kill_liveness_endpoint_process()


def write_config_to_temp_file(tdir_factory: TempdirFactory, data: Text) -> Text:
    temp_dir = tdir_factory.mktemp("consumer-config")
    config_path = temp_dir / "endpoint_config.yml"
    rasa_x_utils.write_text_to_file(config_path, data)

    return config_path


def test_pika_consumer_from_config(
    tmpdir_factory: TempdirFactory,
    monkeypatch: MonkeyPatch,
    event_service: EventService,
):
    # write endpoint config to file
    queue = "q"
    endpoint_config = f"""
event_broker:
    type: "pika"
    url: "url"
    username: "username"
    password: "password"
    queue: "{queue}"
"""

    config_path = write_config_to_temp_file(tmpdir_factory, endpoint_config)

    # patch pika channel initialisation
    from rasa.core.brokers import pika as pika_channel

    monkeypatch.setattr(pika_channel, "initialise_pika_channel", lambda *_: None)

    cfg = read_event_broker_config(config_path)
    actual = consumer_utils.from_endpoint_config(cfg, False)

    assert isinstance(actual, PikaEventConsumer)
    assert actual.queue == queue


@pytest.mark.parametrize("security_protocol", ["plaintext", "ssl"])
def test_kafka_consumer_from_config(
    tmpdir_factory: TempdirFactory, security_protocol: Text
):
    # write endpoint config to file
    url, sasl_username, sasl_password, topic = (
        "some-url",
        "username1",
        "password2",
        "topic3",
    )

    endpoint_config = f"""
event_broker:
    type: kafka
    url: {url}
    sasl_username: {sasl_username}
    sasl_password: {sasl_password}
    topic: {topic}
    security_protocol: {security_protocol}
"""

    config_path = write_config_to_temp_file(tmpdir_factory, endpoint_config)

    cfg = read_event_broker_config(config_path)
    actual = consumer_utils.from_endpoint_config(cfg, False)

    assert isinstance(actual, KafkaEventConsumer)
    assert actual.host == url
    assert actual.sasl_username == sasl_username
    assert actual.sasl_password == sasl_password
    assert actual.topic == topic
    assert actual.security_protocol == security_protocol


def test_kafka_consumer_from_config_with_invalid_security_protocol(
    tmpdir_factory: TempdirFactory,
):
    # write endpoint config to file
    url, sasl_username, sasl_password, topic, security_protocol = (
        "some-url",
        "username1",
        "password2",
        "topic3",
        "unvalid-protocol",
    )

    endpoint_config = f"""
event_broker:
    type: kafka
    url: {url}
    sasl_username: {sasl_username}
    sasl_password: {sasl_password}
    topic: {topic}
    security_protocol: {security_protocol}
"""

    config_path = write_config_to_temp_file(tmpdir_factory, endpoint_config)

    cfg = read_event_broker_config(config_path)
    kafka_consumer = consumer_utils.from_endpoint_config(cfg, False)

    with pytest.raises(ValueError):
        kafka_consumer.consume()


def test_no_endpoint_config_sqlite(tmpdir_factory: TempdirFactory):
    # write empty endpoint config to file
    endpoint_config = ""

    config_path = write_config_to_temp_file(tmpdir_factory, endpoint_config)

    cfg = read_event_broker_config(config_path)

    assert cfg is None
    actual = consumer_utils.from_endpoint_config(cfg, False)
    assert isinstance(actual, SQLiteEventConsumer)


def test_invalid_event_consumer(tmpdir_factory: TempdirFactory):
    # write empty endpoint config to file
    endpoint_config = f"""
event_broker:
    type: invalid_type
"""

    config_path = write_config_to_temp_file(tmpdir_factory, endpoint_config)

    cfg = read_event_broker_config(config_path)

    assert cfg is not None
    with pytest.raises(ValueError):
        consumer_utils.from_endpoint_config(cfg, False)


@pytest.mark.parametrize(
    "endpoint_config,exception",
    [
        ("wrong: section", KeyError),
        ("", TypeError),
        ("random:gibberish:entries", Exception),
    ],
)
def test_read_event_broker_config_exceptions(
    tmpdir_factory: TempdirFactory, endpoint_config: Text, exception
):
    # write endpoint config to file
    config_path = write_config_to_temp_file(tmpdir_factory, endpoint_config)

    with pytest.raises(exception):
        rasa_x_utils.extract_partial_endpoint_config(config_path, "event_broker")


def test_read_event_broker_config_value_error():
    config_path = "/non/existing/path"
    with pytest.raises(ValueError):
        rasa_x_utils.extract_partial_endpoint_config(config_path, "event_broker")


def test_event_service_main_initializes_telemetry(monkeypatch: MonkeyPatch):
    """When the event service is run as a separate service, it should
    initialize_app the telemetry configuration. Otherwise, the config will have
    already been loaded from another process, and the config will be shared
    between that process and the event service's.
    """
    import rasax.community.services.event_service as event_service_module

    initialize_configuration_from_db = Mock()

    monkeypatch.setattr(config, "LOCAL_MODE", False)
    # Some of these functions may have side effects, patch them so that they
    # are not actually run.
    monkeypatch.setattr(event_service_module, "wait_for_migrations", Mock())
    monkeypatch.setattr(event_service_module, "continuously_consume", Mock())
    monkeypatch.setattr(event_service_module, "update_log_level", Mock())
    monkeypatch.setattr(event_service_module, "read_event_broker_config", Mock())
    monkeypatch.setattr(
        telemetry, "initialize_configuration_from_db", initialize_configuration_from_db,
    )

    event_service_module.main(should_run_liveness_endpoint=True)

    initialize_configuration_from_db.assert_called_once()


def test_add_conversation_tags_by_non_existing_id(event_service: EventService):
    conversation_id = generate_conversation(event_service)
    tags = [{"id": 1}]

    with pytest.raises(ValueError):
        event_service.add_conversation_tags(conversation_id, tags)


def test_add_conversation_tags_by_non_existing_conversation(
    event_service: EventService,
):
    with pytest.raises(ValueError):
        event_service.add_conversation_tags("empty", tag_1)


def test_add_conversation_tags_by_value(event_service: EventService,):
    conversation_id = generate_conversation(event_service)

    result = event_service.add_conversation_tags(conversation_id, tag_1)
    assert result[0]["conversations"] == [conversation_id]
    assert result[0]["value"] == tag_1[0]["value"]
    assert result[0]["color"] == tag_1[0]["color"]

    result = event_service.add_conversation_tags(conversation_id, tag_1)
    assert result[0]["conversations"] == [conversation_id]
    assert result[0]["value"] == tag_1[0]["value"]
    assert result[0]["color"] == tag_1[0]["color"]


def test_add_conversation_tags_by_id(event_service: EventService):
    conversation_id = generate_conversation(event_service)
    id = event_service.add_conversation_tags(conversation_id, tag_1)[0]["id"]

    conversation_id = generate_conversation(event_service)
    tag_with_id = [{"id": id}]
    result = event_service.add_conversation_tags(conversation_id, tag_with_id)
    assert len(result[0]["conversations"]) == 2
    assert conversation_id in result[0]["conversations"]


def test_delete_conversation_tags(event_service: EventService):
    conversation_id = generate_conversation(event_service)
    conversation_id_other = generate_conversation(event_service)

    event_service.add_conversation_tags(conversation_id, tag_1)
    tag_id = event_service.add_conversation_tags(conversation_id_other, tag_1)[0]["id"]

    event_service.remove_conversation_tag(conversation_id, tag_id)

    result = event_service.get_conversation_metadata_for_all_clients(
        include_conversation_tags=True
    )[0]

    assert len(get_conversation_by_id(result, conversation_id)["tags"]) == 0
    assert len(get_conversation_by_id(result, conversation_id_other)["tags"]) == 1


async def test_delete_tag_completely_when_not_assigned_to_any_conversation(
    event_service: EventService,
):
    conversation_id = generate_conversation(event_service)
    tag_id = event_service.add_conversation_tags(conversation_id, tag_1)[0]["id"]
    event_service.remove_conversation_tag(conversation_id, tag_id)

    tag_with_id = [{"id": tag_id}]
    with pytest.raises(ValueError):
        event_service.add_conversation_tags(conversation_id, tag_with_id)


async def test_assign_existing_tag_to_another_conversation(event_service: EventService):
    conversation_id = generate_conversation(event_service)
    conversation_id_other = generate_conversation(event_service)

    tag_id = event_service.add_conversation_tags(conversation_id, tag_1)[0]["id"]
    tag_with_id = [{"id": tag_id}]

    result = event_service.add_conversation_tags(conversation_id_other, tag_with_id)
    assert len(result)


async def test_get_all_conversation_tags_empty(event_service: EventService):
    result = event_service.get_all_conversation_tags()
    assert not len(result)


async def test_get_all_conversation_tags(event_service: EventService):
    conversation_id = generate_conversation(event_service)
    conversation_id_other = generate_conversation(event_service)
    event_service.add_conversation_tags(conversation_id, tag_1)
    event_service.add_conversation_tags(conversation_id_other, tag_2)

    result = event_service.get_all_conversation_tags()

    assert len(result) == 2
    assert get_tag_by_value(tag_1[0]["value"], result)["conversations"] == [
        conversation_id
    ]
    assert get_tag_by_value(tag_2[0]["value"], result)["conversations"] == [
        conversation_id_other
    ]
